package org.openjdk.gcbench.roots;

import org.openjdk.jmh.annotations.*;
import org.openjdk.jmh.infra.Blackhole;
import org.openjdk.jmh.infra.Control;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.TimeUnit;

@Warmup(iterations = 5, time = 1, timeUnit = TimeUnit.SECONDS)
@Measurement(iterations = 5, time = 1, timeUnit = TimeUnit.SECONDS)
@Fork(value = 1, jvmArgsAppend = {"-Xss32m"})
@BenchmarkMode(Mode.AverageTime)
@OutputTimeUnit(TimeUnit.MILLISECONDS)
@Threads(Threads.MAX)
@State(Scope.Thread)
public class Synchronizers {

    List<Object> list;

    @Param({"40000"})
    private int size;

    @Setup
    public void setup() {
        list = new ArrayList<>();
        for (int c = 0; c < size; c++) {
            list.add(new Object());
        }
    }

    @Benchmark
    public void test(Control cnt, Blackhole bh) throws InterruptedException {
        recursiveLock(cnt, bh, list, 0);
    }

    private void recursiveLock(Control cnt, Blackhole bh, List<Object> list, int i) {
        if (i < list.size()) {
            Object o0 = list.get(i + 0);
            Object o1 = list.get(i + 1);
            Object o2 = list.get(i + 2);
            Object o3 = list.get(i + 3);
            Object o4 = list.get(i + 4);
            Object o5 = list.get(i + 5);
            Object o6 = list.get(i + 6);
            Object o7 = list.get(i + 7);
            Object o8 = list.get(i + 8);
            Object o9 = list.get(i + 9);
            synchronized (o0) {
                synchronized (o1) {
                    synchronized (o2) {
                        synchronized (o3) {
                            synchronized (o4) {
                                synchronized (o5) {
                                    synchronized (o6) {
                                        synchronized (o7) {
                                            synchronized (o8) {
                                                synchronized (o9) {
                                                    recursiveLock(cnt, bh, list, i + 10);
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        } else {
            for (Object o : list) {
                bh.consume(o.hashCode());
            }
            if (cnt.startMeasurement) {
                // Do not leave until we are finished
                while (!cnt.stopMeasurement) {
                    bh.consume(new Object());
                }
            }
        }
    }

}
